﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2018 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis {
  [Item("Transformed Regression Model", "A model that was transformed back to match the original variables after the training was performed on transformed variables.")]
  [StorableClass]
  public class TransformedRegressionModel : RegressionModel, ITransformedRegressionModel {

    [Storable]
    public IRegressionModel OriginalModel { get; private set; }

    [Storable]
    public ItemList<IDataAnalysisTransformation> Transformations { get; private set; }

    public override IEnumerable<string> VariablesUsedForPrediction {
      get { return OriginalModel.VariablesUsedForPrediction; }
    }

    #region Constructor, Cloning & Persistence
    public TransformedRegressionModel(IRegressionModel originalModel, IEnumerable<IDataAnalysisTransformation> transformations)
      : base(RegressionProblemData.GetOriginalTragetVariable(originalModel.TargetVariable, transformations)) {
      Name = "Transformed " + originalModel.Name;
      OriginalModel = originalModel;
      Transformations = new ItemList<IDataAnalysisTransformation>(transformations);
    }

    protected TransformedRegressionModel(TransformedRegressionModel original, Cloner cloner)
      : base(original, cloner) {
      OriginalModel = cloner.Clone(original.OriginalModel);
      Transformations = cloner.Clone(original.Transformations);
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new TransformedRegressionModel(this, cloner);
    }

    [StorableConstructor]
    protected TransformedRegressionModel(bool deserializing)
      : base(deserializing) { }
    #endregion

    // dataset in original data range
    public override IEnumerable<double> GetEstimatedValues(IDataset dataset, IEnumerable<int> rows) {
      var transformedDataset = TransformInputs(dataset, Transformations);

      var estimates = OriginalModel.GetEstimatedValues(transformedDataset, rows);

      return InverseTransformEstimates(estimates, Transformations, OriginalModel.TargetVariable);
    }

    // problemData in original data range
    public override IRegressionSolution CreateRegressionSolution(IRegressionProblemData problemData) {
      return new TransformedRegressionSolution(this, new RegressionProblemData(problemData));
    }

    private static IDataset TransformInputs(IDataset dataset, IEnumerable<IDataAnalysisTransformation> transformations) {
      return DataAnalysisProblemData.Transform(dataset, transformations);
    }

    private static IEnumerable<double> InverseTransformEstimates(IEnumerable<double> data, IEnumerable<IDataAnalysisTransformation> transformations, string targetVariable) {
      var estimates = data.ToList();

      foreach (var transformation in transformations.Reverse()) {
        if (transformation.TransformedVariable == targetVariable) {
          var trans = (ITransformation<double>)transformation.Transformation;

          estimates = trans.InverseApply(estimates).ToList();

          // setup next iteration
          targetVariable = transformation.OriginalVariable;
        }
      }

      return estimates;
    }
  }
}